/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */
package de.bsvrz.ste.rdstmc.meldungserstellung;
 
import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.dav.daf.main.Data.NumberArray;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.ste.rdstmc.allgemein.RDSEreignis;
import de.bsvrz.ste.rdstmc.konfiguration.TmcLocationPunkt;
import de.bsvrz.sys.funclib.dambach.parameter.ParameterSender;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Die Klasse zum Versenden von Daten unter dem Aspekt "Senden" der
 * Attributgruppe "RDSMeldung". 
 * Sie implementiert das ClientSenderInterface welches die Methoden
 * zum Versenden der Daten bereitstellt.
 * 
 * @author Dambach Werke GmbH
 * @version $Revision: 1.3 $ / $Date: 2009/01/27 13:54:22 $ / ($Author: Sans $)
 * 
 */
public class Meldung implements ClientSenderInterface
{
  /**
   * Debug-Logger für Logging-Ausgaben
   */
  private final static Debug _debug = Debug.getLogger();

  /**
   * Ist auf true gesetzt, solange auf Sendesteuerung gewartet wird
   */
  private boolean _warten = false;

  /**
   * Rolle, mit der als Sender angemeldet wird
   */
  public static final SenderRole ROLLE = SenderRole.source();

  /**
   * Zeitin ms, die max. auf Sendesteuerung gewartet wird
   */
  public static final long SENDE_TIMEOUT = 5000;

  /**
   * Ist diese Instanz zum DAV-Senden angemeldet?
   */
  private boolean _angemeldet = false;

  /**
   * @return ist diese Instanz zum DAV-Senden angemeldet?
   */
  public boolean isAngemeldet()
  {
    return _angemeldet;
  }

  private VerwaltungMeldungen _verwaltung;

  protected VerwaltungMeldungen getVerwaltung()
  {
    return _verwaltung;
  }

  /**
   * @return die Datenverteiler-Verbindung
   */
  public ClientDavInterface getConnection()
  {
    return getVerwaltung().getConnection();
  }

  private SystemObject _davObjekt;

  /**
   * @return das Dav-Systemobject, für das das Senden von Daten erfolgt
   */
  public SystemObject getDavObjekt()
  {
    return _davObjekt;
  }

	/**
	 * Konstruktor der Klasse
	 *
	 * @param verwaltung VerwaltungMeldungen
	 * @param davObjekt SystemObject
	 * @param datenGeneriert ResultData
	 */
  public Meldung(VerwaltungMeldungen verwaltung, SystemObject davObjekt, ResultData datenGeneriert)
  {
    _verwaltung = verwaltung;
    _davObjekt = davObjekt;
    _datenGeneriert = datenGeneriert;

    if (datenGeneriert != null)
    {
      _version = getVersion(_datenGeneriert.getData());
      _erzeugungsZeit = getErzeugungsZeit(_datenGeneriert.getData());
    }

    try
    {
      getConnection().subscribeSender(this, getDavObjekt(), getVerwaltung().getDatenBeschreibung(), ROLLE);
      _debug.finer(getVerwaltung().getDatenBeschreibung() + " zum Senden angemeldet für " + getDavObjekt());
      _angemeldet = true;
    }
    catch (OneSubscriptionPerSendData e)
    {
      /* Es liegt bereits eine Sendeanmeldung von anderer Stelle vor
       * => hier nicht anmnelden
       */
      _debug.warning(getDavObjekt().getPid() +
          " ist bereits an anderer Stelle zum Senden angemeldet", e);
    }
  }

  public Meldung(VerwaltungMeldungen verwaltung, SystemObject davObjekt)
  {
    this(verwaltung, davObjekt, null);
  }

  /**
   * Gibt die mit dieser Instanz verbundenen Ressourcen wieder frei
   */
  public void dispose()
  {
    if (_angemeldet)
    {
      getConnection().unsubscribeSender(this, getDavObjekt(), getVerwaltung().getDatenBeschreibung());
      _debug.config("Senden abgemeldet von " + getVerwaltung().getDatenBeschreibung() + " für " + getDavObjekt());
      _angemeldet = false;
    }
  }

  protected boolean send(Ereignis ereignis)
  {
    ResultData resultData = getResultData(ereignis);
    if (send(resultData))
    {
      new ParameterSender(getConnection(), getDavObjekt(), resultData, "atg.rdsMeldungGeneriert").start();
      _datenGeneriert = resultData;
      _version = getVersion(_datenGeneriert.getData());
      _erzeugungsZeit = getErzeugungsZeit(_datenGeneriert.getData());

      return true;
    }
    else
    {
      return false;
    }
  }

  /**
   * Unmittelbares Senden von Daten.
   *
   * @param resultData die zu sendenden Daten
   * @return true, wenn das Senden erfolgreich war, false sonst
   */
  boolean send(ResultData resultData)
  {
    try
    {
      getConnection().sendData(resultData);
    }
    catch(SendSubscriptionNotConfirmed e)
    {
      if (SENDE_TIMEOUT > 0)
      {
        _debug.config("Warte max. " + SENDE_TIMEOUT + "ms auf Sendesteuerung für " + getDavObjekt().getPid());
        synchronized(this)
        {
          _warten = true;
          try
          {
            wait(SENDE_TIMEOUT);
          }
          catch(InterruptedException ex)
          {
            // tue nichts
          }
          _warten = false;
        }
        try
        {
          getConnection().sendData(resultData);
        }
        catch (SendSubscriptionNotConfirmed ex)
        {
          _debug.warning("Keine Sendesteuerung für " + getDavObjekt().getPid(), ex);
          return false;
        }
        return true;
      }
      else
      {
        _debug.warning("Keine Sendesteuerung für " + getDavObjekt().getPid(), e);
        return false;
      }
    }
    catch (DataNotSubscribedException e)
    {
      _debug.error("Keine Sendeanmeldung bzw. Sendeanmeldung ungültig für " + getDavObjekt().getPid(), e);
      return false;
    }
    return true;
  }

  /*
   * (Kein Javadoc)
   * @see de.bsvrz.dav.daf.main.ClientSenderInterface#dataRequest(de.bsvrz.dav.daf.main.config.SystemObject, de.bsvrz.dav.daf.main.DataDescription, byte)
   */
  public void dataRequest(SystemObject object, DataDescription dataDescription, byte state)
  {
    _debug.finer("dataRequest() für " + getDavObjekt().getPid() + " - Sender state: " + state);
    if (_warten && state == ClientSenderInterface.START_SENDING)
    {
      synchronized(this)
      {
        notify();
      }
    }
  }

  /*
   * (Kein Javadoc)
   * @see de.bsvrz.dav.daf.main.ClientSenderInterface#isRequestSupported(de.bsvrz.dav.daf.main.config.SystemObject, de.bsvrz.dav.daf.main.DataDescription)
   */
  public boolean isRequestSupported(SystemObject so, DataDescription dd)
  {
    return true;
  }

  /**
   * Aktuell unter dem Aspekt generiert gesendete Daten
   */
  private ResultData _datenGeneriert = null;

  /**
   * @return die unter dem Aspekt senden aktuell gesendeten Daten
   */
  protected ResultData getDatenGeneriert()
  {
    return _datenGeneriert;
  }

  /**
   * setzt die unter dem Aspekt senden aktuell gesendeten Daten
   *
   * @param datenSenden die aktuell gesendeten Daten
   */
//  void setDatenSenden(ResultData datenSenden)
//  {
//    _datenGeneriert = datenSenden;
//  }

  int _version = 0;
  long _erzeugungsZeit = 0;


  /**
   * Bildet eine sendefertige ResultData-Instanz für das aktuelle Meldungsobjekt mit dem Zeitstempel der aktuellen Zeit
   * und einem DAV-Datensatz für das übergebene Ereignis
   *
   * @param ereignis das Ereignis, für das der die Daten erzeugt werden sollen
   *
   * @return ResultData Instanz
   */
  private ResultData getResultData(Ereignis ereignis)
  {
    Data data = getData(ereignis);
    AttributeGroup atg = getConnection().getDataModel().getAttributeGroup("atg.rdsMeldung");
    Aspect asp = getConnection().getDataModel().getAspect("asp.rdsGeneriert");
    return new ResultData(getDavObjekt(), new DataDescription(atg, asp), getConnection().getTime(), data);
  }

  /**
   * Provisorischer Wert für "undefinierte" Zeitangabe
   */
  public final static long NULLZEIT = 31532400000L;

  /**
   * Bildet einen DAV-Datensatz zur vorliegenden Meldung und initialisiert die relevanten Attribute bezg.
   * Location(s) und Ereignis
   *
   * @param ereignis das Ereignis, für das der Datensatz gebildet werden soll
   *
   * @return den gebildeten DAV-Datensatz
   */
  private Data getData(Ereignis ereignis)
  {
    long aktuelleZeit = getConnection().getTime();
    long erzeugungsZeit = _version == 0 ? aktuelleZeit : _erzeugungsZeit;
    long ablaufZeit = getVerwaltung().getAblaufOffset() > 0 ?
        aktuelleZeit + getVerwaltung().getAblaufOffset() * 60 * 1000 : NULLZEIT;
    int version = _version + 1;

    AttributeGroup atg = getConnection().getDataModel().getAttributeGroup("atg.rdsMeldung");
    Data data = getConnection().createData(atg);

    data.getItem("InfoGUID").asTextValue().setText("" + getDavObjekt().getId());
    Data dataId = data.getItem("ID");

    dataId.getItem("IDOrganisation").asTextValue().setText(getVerwaltung().getIdOrganisation() == null ?
        "" : getVerwaltung().getIdOrganisation());
    dataId.getItem("IDNummer").asTextValue().setText(getVerwaltung().getIdNummer() == null ?
        "" : getVerwaltung().getIdNummer());

    Data dataVersion = data.getItem("Version");
    dataVersion.getItem("VersionGUID").asTextValue().setText("" + getDavObjekt().getId() + "-" + version);

    Data dataVerwaltung = dataVersion.getItem("VerwaltungsInformationen");
    dataVerwaltung.getItem("AuthorisierendeOrganisationsEinheit").asTextValue().setText("");
    dataVerwaltung.getItem("AuthorisierenderNutzer").asTextValue().setText("");

    dataVerwaltung.getItem("AuthorisierungsZeit").asTimeValue().setMillis(NULLZEIT);
    dataVerwaltung.getItem("AuthorisierungsKommentar").asTextValue().setText("");
    // dataVerwaltung.getItem("AuthorisierungsErgebnis").asUnscaledValue().set(0);
    dataVerwaltung.getItem("VersionsNummer").asUnscaledValue().set(version);

    dataVerwaltung.getItem("AktualisierungsZeit").asTimeValue().setMillis(aktuelleZeit);
    dataVerwaltung.getItem("AktivierungsZeit").asTimeValue().setMillis(erzeugungsZeit);
    dataVerwaltung.getItem("AblaufZeit").asTimeValue().setMillis(ereignis.getRDSEreignis() == RDSEreignis.KEINE_AUSSAGE ?
          aktuelleZeit : ablaufZeit);
    dataVerwaltung.getItem("ErinnerungsZeit").asTimeValue().setMillis(NULLZEIT);
    dataVerwaltung.getItem("ErinnerungsZeitOffset").asTimeValue().setMillis(0);
    dataVerwaltung.getItem("ErinnerungsTyp").asUnscaledValue().set(0);
    dataVerwaltung.getItem("ErzeugungsZeit").asTimeValue().setMillis(erzeugungsZeit);
    Data dataAenderungsInfo = dataVerwaltung.getItem("ÄnderungsInformationen");
    dataAenderungsInfo.getItem("Veranlasser").asTextValue().setText("SYSTEM");
    dataAenderungsInfo.getItem("ModifikationsZeitpunkt").asTimeValue().setMillis(aktuelleZeit);
    dataAenderungsInfo.getItem("ModifikationsKommentar").asTextValue().setText("");
    dataVerwaltung.getItem("Status").asTextValue().setText(ereignis.getRDSEreignis() == RDSEreignis.KEINE_AUSSAGE ?
        "Aufhebung" : "Modifikation");
    // dataVerwaltung.getItem("Wichtung").asUnscaledValue().set(0);
    dataVerwaltung.getItem("Priorität").asUnscaledValue().set(0); // 0 == "Normal"
    dataVerwaltung.getItem("LandesKennung").asTextValue().setText("rdsLandesKennung.DE");
    dataVerwaltung.getItem("Zustand").asTextValue().setText("nicht quittiert");
    dataVerwaltung.getItem("Quelle").asTextValue().setText("automatisch");

    Data dataVerkehr = dataVersion.getItem("VerkehrsInformationen");
    dataVerkehr.getItem("RDSFormat").asUnscaledValue().set(0);

    Data dataNachrichten = dataVerkehr.getItem("Nachrichten");
    // Data dataNachrichtenDaten = dataVerkehr.getItem("NachrichtenDaten");
    dataNachrichten.getItem("NachrichtenSprache").asUnscaledValue().set(1031);

    Data dataFreierText = dataVerkehr.getItem("FreierText");
    String freiText = "";
//    SystemObject ek = ereignis.getErklaerungsKomponente();
//    if (ek != null)
//    {
//      freiText = "Referenzierte Störungsursache: " + ek.getName() + " (pid: " + ek.getPid() + ")";
//    }
    dataFreierText.getItem("Text").asTextValue().setText(freiText);

    dataFreierText.getItem("TextSprache").asUnscaledValue().set(1031);

    Data dataLocationTabelleInfo = dataVerkehr.getItem("LocationTabelleInfo");
    dataLocationTabelleInfo.getItem("LocationTabelle").asUnscaledValue().set(1001);
    dataLocationTabelleInfo.getItem("LocationTabelleVersion").asTextValue().setText("");
    dataLocationTabelleInfo.getItem("LocationTabelleTyp").asUnscaledValue().set(0);
    dataLocationTabelleInfo.getItem("LocationTabelleName").asTextValue().setText("");

    Data dataLocationDaten = dataVerkehr.getItem("LocationDaten");
    Array dataRDSLocation = dataLocationDaten.getArray("RDSLocation"); // !!! Location Codes

    switch (getVerwaltung().getLocationMethode())
    {
      case LISTE:
        dataRDSLocation.setLength(ereignis.getTmcOrte().size());
        break;
      case EXTEND:
        dataRDSLocation.setLength(1);
        break;
      default:
        throw new UnsupportedOperationException("Location-Methode mit Id \"" +
            getVerwaltung().getLocationMethode().id + "\" nicht unterstützt");
    }

    for (int i = 0; i < dataRDSLocation.getLength(); i +=1 )
    {
      Data locationDaten = dataRDSLocation.getItem(i);
      TmcLocationPunkt tmcPunkt = ereignis.getTmcOrte().get(i);
      locationDaten.getItem("LocationCode").asUnscaledValue().set(tmcPunkt.getLocationCode());
      locationDaten.getItem("StrassenNummer").asTextValue().setText(tmcPunkt.getStrassenNummer());
      locationDaten.getItem("AusfahrtNummer").asTextValue().setText(tmcPunkt.getAusfahrtNummer());
      locationDaten.getItem("TmcOrtsTyp").asUnscaledValue().set(tmcPunkt.getOrtsTyp());
      Data koordinaten = locationDaten.getItem("Koordinaten");
      koordinaten.getItem("RDSLocationKoordinateX").asScaledValue().set(tmcPunkt.getX());
      koordinaten.getItem("RDSLocationKoordinateY").asScaledValue().set(tmcPunkt.getY());
      // locationDaten.getItem("LocationKilometrierung").asTextValue().setText("0");
      locationDaten.getItem("LocationCodeLinienReferenz").asUnscaledValue().set(tmcPunkt.getLinienCode());
      locationDaten.getItem("LocationCodeGebietsReferenz").asUnscaledValue().set(tmcPunkt.getGebietsCode());
      locationDaten.getItem("LocationNextNegativ").asUnscaledValue().set(tmcPunkt.getVorgaengerCode());
      locationDaten.getItem("LocationNextPositiv").asUnscaledValue().set(tmcPunkt.getNachfolgerCode());
      Data locationNachricht = locationDaten.getItem("LocationNachricht");
      // Array nachrichtenDaten = locationNachricht.getArray("NachrichtenDaten");
      locationNachricht.getItem("NachrichtenSprache").asUnscaledValue().set(1031);
    }

    // dataLocationDaten.getItem("RDSLocationFormat")
    dataLocationDaten.getItem("LocationExtent").asUnscaledValue().set(ereignis.getAusdehnung());
    dataLocationDaten.getItem("LocationMethode").asUnscaledValue().set(getVerwaltung().getLocationMethode().id);
    dataLocationDaten.getItem("TMCRichtung").asUnscaledValue().set(ereignis.getTmcRichtung());
    dataLocationDaten.getItem("LocationPrimärEntfernung").asUnscaledValue().set(0);
    dataLocationDaten.getItem("LocationSekundärEntfernung").asUnscaledValue().set(0);

    TmcLocationPunkt sekLocation = ereignis.getTmcOrte().get(ereignis.getTmcOrte().size() - 1);
    int locationVor = sekLocation.getVorgaenger() != null ? sekLocation.getVorgaenger().getLocationCode() : 0;
    dataLocationDaten.getItem("LocationVorLocation").asUnscaledValue().set(locationVor);

    TmcLocationPunkt primLocation = ereignis.getTmcOrte().get(0);
    int locationNach = primLocation.getNachfolger() != null ? primLocation.getNachfolger().getLocationCode() : 0;
    dataLocationDaten.getItem("LocationNachLocation").asUnscaledValue().set(locationNach);
    // dataLocationDaten.getItem("LocationKategorie").asUnscaledValue().set(0);
    // dataLocationDaten.getItem("LocationRichtungTextID").asUnscaledValue().set(0);

    Data dataEreignis = dataVerkehr.getItem("Ereignis");

    Array dataEreignisDaten = dataEreignis.getArray("EreignisDaten"); // !!! Ereignis Codes
    dataEreignisDaten.setLength(1);
    Data ereignisDaten = dataEreignisDaten.getItem(0);
    ereignisDaten.getItem("EreignisKategorie").asUnscaledValue().set(1);
    ereignisDaten.getItem("EreignisCode").asTextValue().setText(ereignis.getEreignisCode());
    ereignisDaten.getItem("EmpfehlungsCode").asUnscaledValue().set(0);
    ereignisDaten.getItem("VorhersageCode").asUnscaledValue().set(0);
    ereignisDaten.getItem("EreignisDauer").asUnscaledValue().set(0);
    Array dataEreignisQuantitaet = ereignisDaten.getArray("EreignisQuantität");
    RDSQuantitaetLEN rdsQuantitaetLEN = ereignis.getRDSQuantitaetLEN();
    if (rdsQuantitaetLEN != null)
    {
      dataEreignisQuantitaet.setLength(1);
      Data quantitaetDatenDaten = dataEreignisQuantitaet.getItem(0);
      quantitaetDatenDaten.getItem("QuantitätsKennung").asTextValue().setText(RDSQuantitaetLEN.KENNNUNG);
      quantitaetDatenDaten.getItem("QuantitätsWert").asTextValue().setText("" + rdsQuantitaetLEN.getBetroffeneLaenge());
      quantitaetDatenDaten.getItem("QuantitätsEinheit").asTextValue().setText(RDSQuantitaetLEN.EINHEIT);
    }
    
    // dataEreignis.getItem("EreignisInBeidenRichtungen").asTextValue().setText("Nein");
    // dataEreignis.getItem("EreignisKodierung").asUnscaledValue().set(0);
    dataEreignis.getItem("EreignisTabelleNummer").asUnscaledValue().set(0);
    dataEreignis.getItem("EreignisTabelleVersion").asTextValue().setText("");
    NumberArray arrayEreignisTyp = dataEreignis.getUnscaledArray("EreignisTyp");
    arrayEreignisTyp.setLength(1);
    arrayEreignisTyp.getValue(0).set(0);
    
    // Data dataZusatzAttribute = dataVerkehr.getItem("ZusatzAttribute");
    // dataZusatzAttribute.getItem("GeographischeRelevanz").asUnscaledValue().set(3);
    // dataZusatzAttribute.getItem("Vorhersage").asTextValue().setText("Nein");
    
    dataVerkehr.getItem("AlertCCode").asTextValue().setText(ereignis.getAlertCCode());
    
    return data;
  }
   
  private static int getVersion(Data data)
  {
    Data dataVersion = data.getItem("Version");
    Data dataVerwaltung = dataVersion.getItem("VerwaltungsInformationen");
    return dataVerwaltung.getItem("VersionsNummer").asUnscaledValue().intValue();
  }  
  
  private static long getErzeugungsZeit(Data data)
  {
    Data dataVersion = data.getItem("Version");
    Data dataVerwaltung = dataVersion.getItem("VerwaltungsInformationen");
    return dataVerwaltung.getItem("ErzeugungsZeit").asTimeValue().getMillis();
  }
    
}
